import base64
import tempfile
import compressed_rtf
from email_master.util import EmailAttachmentList, EmailAttachment, EmailParser
from email_master.compat import to_unicode
from extract_msg import Message
from extract_msg.utils import xstr
import hashlib
import re
import six

if six.PY3:
    unicode = to_unicode
    UnicodeType = str
else:
    UnicodeType = unicode


class MSGMessageObj(Message, object):
    def __init__(self, *args, **kwargs):
        self._getStringStream = self.fixed_ss
        Message.__init__(self, *args, **kwargs)  # Old class-style super().__init__ call

    @property
    def stringEncoding(self):
        """
        PATCHED stringEncoding, to allow failing encodings to resolve using utf-8
        """

        try:
            return self.__stringEncoding
        except AttributeError:
            # We need to calculate the encoding
            # Let's first check if the encoding will be unicode:
            if self.areStringsUnicode:
                self.__stringEncoding = "utf-16-le"
                return self.__stringEncoding
            else:
                # Well, it's not unicode. Now we have to figure out what it IS.
                if not self.mainProperties.has_key('3FFD0003'):
                    self.__stringEncoding = "utf-8"  # Unable to determine encoding, try catch-all utf-8
                    return self.__stringEncoding
                    # raise Exception('Encoding property not found')

                enc = self.mainProperties['3FFD0003'].value
                # Now we just need to translate that value
                # Now, this next line SHOULD work, but it is possible that it might not...
                self.__stringEncoding = str(enc)
                return self.__stringEncoding

    @property
    def body(self):
        """
        PATCHED func from lib for MSG encoding issues

        Returns the message body, if it exists.
        """
        try:
            return self._body
        except AttributeError:
            self._body = self._getStringStream('__substg1.0_1000')
            if self._body:
                a = re.search('\n', self._body)
                if a is not None:
                    if re.search('\r\n', self._body) is not None:
                        self.__crlf = '\r\n'
            return self._body

    def fixed_ss(self, filename, prefix=True):
        """
        __getStringStream()

        PATCHED func for MSG encoding issues
        """
        from extract_msg.utils import windowsUnicode
        filename = self.fix_path(filename, prefix)
        if self.areStringsUnicode:
            return windowsUnicode(self._getStream(filename + '001F', prefix=False))
        else:
            tmp = self._getStream(filename + '001E', prefix=False)
            if not tmp:
                return None

            try:
                tmp = tmp.decode(self.stringEncoding)  # The MSG lied about string encoding, we need to figure it out
            except UnicodeDecodeError:
                tmp = tmp.decode("utf-8", errors="replace")  # Last attempt at decoding before exception

            return tmp


class MSGParser(EmailParser):
    """
    See https://msdn.microsoft.com/en-us/library/cc433490(v=exchg.80).aspx for full property list
    """

    def __init__(self, email_data, filename="unknown.msg", ignore_errors=False, exclude_attachment_extensions=None,
                 include_attachment_extensions=None):
        """
        Parse an MSG File
        :param fn: filename of msg file
        :param email_data: base64 data of the MSG file
        """
        super(MSGParser, self).__init__(email_data,
                                        filename=filename,
                                        ignore_errors=ignore_errors,
                                        exclude_attachment_extensions=exclude_attachment_extensions,
                                        include_attachment_extensions=include_attachment_extensions)
        temp_file = tempfile.SpooledTemporaryFile()  # Create a temp file for the ole library
        temp_file.write(base64.b64decode(email_data))

        self.msg = MSGMessageObj(temp_file)

    def _get_msg_data(self, filename):
        if self.msg.areStringsUnicode:
            charset = "utf_16_le"
            suffix = "001F"
        else:
            charset = self.msg.stringEncoding
            suffix = "001E"

        result = self.msg._getStream(filename + suffix, prefix=False)

        if result is None:
            return u""

        try:
            result = result.decode(charset)
        except Exception:
            # Fall back to UTF-8 (If we fail here it's unparsable and should throw an error)
            result = result.decode("utf-8")

        return result

    def get_cc(self):
        return self.msg.cc or u""

    def get_bcc(self):
        return self._get_msg_data('__substg1.0_0E02')

    def get_raw_content(self):
        return self.email_data

    def get_raw_headers(self):
        return u"\n".join([u"{}: {}".format(h[0], h[1]) for h in self.get_headers() or []])

    def get_reply_to(self):
        return self._get_msg_data('__substg1.0_1042')

    def get_id(self):
        return self._get_msg_data('__substg1.0_1035')

    def get_sender(self):
        return self.msg.sender or u""

    def get_plaintext_body(self):
        return self.msg.body or u""

    def get_html_body(self, decode_html=True):
        return self._get_msg_data('__substg1.0_1013')

    def get_rtf_body(self):
        data = self._get_msg_data("__substg1.0_10090102")
        if data:
            return compressed_rtf.decompress(data) or u""
        else:
            return u""

    def get_date(self):
        return self.msg.date or u""

    def get_subject(self):
        return self.msg.subject or u"(No Subject)"

    def get_type(self):
        return u"MSG"

    def get_recipients(self):
        return u",".join([self.msg.cc or u"", self.msg.to or u""])

    def get_headers(self):
        headers = self.msg.header._headers
        return headers if headers else []

    def get_attachments(self):
        raw_attachments = self.msg.attachments
        attachments = EmailAttachmentList()

        for raw_attachment in raw_attachments:
            if raw_attachment.data:
                data = raw_attachment.data
            else:
                data = self.msg._getStream(raw_attachment.dir + '__substg1.0_37010102')

            if not data:
                # No data in the attachment ignore
                continue
            elif isinstance(data, Message):
                # Following code taken from extract_msg.save()
                eml_data = 'From: ' + xstr(self.msg.sender) + "\n"
                eml_data += 'To: ' + xstr(self.msg.to) + "\n"
                eml_data += 'CC: ' + xstr(self.msg.cc) + "\n"
                eml_data += 'Subject: ' + xstr(self.msg.subject) + "\n"
                eml_data += 'Date: ' + xstr(self.msg.date) + "\n"
                eml_data += '-----------------' + "\n" + "\n"
                eml_data += self.msg.body

                data = eml_data
            if isinstance(data, UnicodeType):
                data = data.encode("utf-8", errors="replace")

            fallback_name = u"unknown-{}".format(hashlib.md5(data).hexdigest())

            filename = raw_attachment.longFilename or fallback_name
            header_info = raw_attachment.shortFilename or fallback_name

            attachments.add_attachment(EmailAttachment(header_info, filename, data))

        return attachments



from email_master.messages import EmailMasterMessage
from email_master.util import EmailUtil
from email_master.pgp import CustomMIMEWrapper, PGPConfig
from email_master.compat import base64_encode, to_unicode
from email_master.compat import MIMEMultipart, EmailMessage
from email.utils import make_msgid, formatdate
import mimetypes
import json
import base64
import smtplib
import ssl
import six


if six.PY2:
    from email.MIMEText import MIMEText
elif six.PY3:
    StringTypes = str
    unicode = to_unicode


class EMLMessage(EmailMasterMessage):
    def __init__(self, pgp_config=None, debug=False):
        super(EMLMessage, self).__init__(EmailMessage(), pgp_config)
        self._debug = debug
        self._debug_info = {}

    def _get_to_addrs(self):
        return list(filter(lambda x: x, [self.message_obj["To"], self.message_obj["Cc"], self.message_obj["Bcc"]]))

    def test_conn(self, username, password, host, port, verify_conn=True):
        conn = smtplib.SMTP(host, port, timeout=5)
        conn.ehlo()
        if verify_conn:
            context = ssl.create_default_context()
        else:
            context = ssl._create_unverified_context()

        if six.PY3:
            conn.starttls(context=context)
        else:
            conn.starttls()
            conn.ehlo()
        if password != EmailMasterMessage.NO_PASSWORD:  # Don't log in if no password
            conn.login(username, password)

        conn.noop()  # test connection
        return conn

    def send(self, username, password, host, port, verify_conn=True, attach_hook=None):
        self.message_obj["Message-ID"] = make_msgid()
        self.message_obj["Date"] = formatdate()

        for attch in self._attachments.attachments:  # Add attachments before sending
            fn = attch.filename
            data = attch.raw_data

            ctype_and_encoding = mimetypes.guess_extension(fn)
            if ctype_and_encoding is not None and len(ctype_and_encoding) == 2:
                ctype, encoding = ctype_and_encoding
            else:
                ctype = None
                encoding = None

            if ctype is None or encoding is not None:
                ctype = "application/octet-stream"
            maintype, subtype = ctype.split("/", 1)
            if six.PY3:
                if attach_hook:
                    attch_kwargs = attach_hook(attch)
                else:
                    attch_kwargs = {
                        "maintype": maintype,
                        "subtype": subtype,
                        "filename": fn
                    }
                self.message_obj.add_attachment(
                    data,
                    **attch_kwargs
                )
            else:
                attch_payload = MIMEText(base64_encode(data) + "\r\n", "base64")
                attch_payload.set_charset("utf-8")
                attch_payload.add_header("Content-Disposition", "attachment", filename=fn)
                attch_payload.add_header("Content-Transfer-Encoding", "base64")
                self.message_obj._payload[0].attach(attch_payload)

        if self._debug:
            self._debug_info["pre_sign"] = base64.b64encode(self.message_obj.as_string().encode()).decode()

        signed_payload = self.pgp_config.lock(CustomMIMEWrapper(self.message_obj))
        if self._debug:
            self._debug_info["post_sign_pre_send"] = base64.b64encode(signed_payload.encode()).decode()

        conn = self.test_conn(username, password, host, port, verify_conn)
        if not self._debug:
            # Only send if we're not in debug mode
            conn.sendmail(EmailUtil.text_to_encoded_words(self.message_obj["From"]), self._get_to_addrs(), signed_payload)

    def get_summary(self):
        data = {
            "message_type": str(self.pgp_config.pgp_action.to_ingest_type()),
            "sent_to": ",".join(self._get_to_addrs()),
            "num_attachments": len(self._attachments.attachments),
            "attach_filenames": ",".join([a.filename for a in self._attachments.attachments]),
            "attach_hashes_md5": ",".join([a.hash_md5 for a in self._attachments.attachments]),
            "attach_hashes_sha1": ",".join([a.hash_sha256 for a in self._attachments.attachments]),
            "attach_hashes_sha256": ",".join([a.hash_sha256 for a in self._attachments.attachments]),
            "subject": self.message_obj["Subject"]
        }

        if self._debug:
            data["debug_info"] = base64.b64encode(json.dumps(self._debug_info).encode()).decode()

        return data



import six
from email.encoders import encode_7or8bit
from email.mime.application import MIMEApplication
from email.utils import collapse_rfc2231_value
from email import message_from_string
import email_master.rfc3156 as rfc3156
import copy
import re
from pgpy import PGPMessage
from pgpy.constants import SymmetricKeyAlgorithm, HashAlgorithm


if six.PY3:
    from io import StringIO
else:
    from cStringIO import StringIO


class CustomMIMEWrapper(object):
    """PGP/MIME (RFC1847 + RFC3156) compliant wrapper."""
    _signature_subtype = 'pgp-signature'
    _encryption_subtype = 'pgp-encrypted'
    _keys_subtype = 'pgp-keys'
    _signed_type = 'application/' + _signature_subtype
    _encrypted_type = 'application/' + _encryption_subtype
    _keys_type = 'application/' + _keys_subtype
    _signed_multipart = 'multipart/signed'
    _encrypted_multipart = 'multipart/encrypted'
    _signature_preamble = \
        'This is an OpenPGP/MIME signed message (RFC 4880 and 3156)'
    _encryption_preamble = \
        'This is an OpenPGP/MIME encrypted message (RFC 4880 and 3156)'

    def __init__(self, msg):
        self.msg = msg

    def get_payload(self):
        yield self.msg.as_string()

    def _is_mime(self):
        is_multipart = self.msg.is_multipart()
        payloads = len(self.msg.get_payload()) if self.msg.get_payload() else 0
        return is_multipart and payloads == 2

    def _micalg(self, hash_algo):
        algs = {
            HashAlgorithm.MD5: 'md5',
            HashAlgorithm.SHA1: 'sha1',
            HashAlgorithm.RIPEMD160: 'ripemd160',
            HashAlgorithm.SHA256: 'sha256',
            HashAlgorithm.SHA384: 'sha384',
            HashAlgorithm.SHA512: 'sha512',
            HashAlgorithm.SHA224: 'sha224'
        }
        return 'pgp-' + algs[hash_algo]

    def openpgp_mangle_for_signature(self, msg):
        """Return a message suitable for signing.
        Encodes multipart message parts in msg as base64, then renders the
        message to string enforcing the right newline conventions. The
        returned value is suitable for signing according to RFC 3156.
        The incoming message is modified in-place.
        """
        rfc3156.encode_base64_rec(msg)
        fp = StringIO()
        g = rfc3156.RFC3156CompliantGenerator(
            fp, mangle_from_=False, maxheaderlen=76)

        g.flatten(msg)

        s = re.sub('\r?\n', '\r\n', fp.getvalue())
        if msg.is_multipart():
            if not s.endswith('\r\n'):
                s += '\r\n'
        return s

    def _wrap_signed(self, msg, signature):
        self.msg.set_payload([])
        self.msg.attach(msg)
        self.msg.set_type(CustomMIMEWrapper._signed_multipart)
        self.msg.set_param('micalg', self._micalg(signature.hash_algorithm))
        self.msg.set_param('protocol', CustomMIMEWrapper._signed_type)
        self.msg.preamble = CustomMIMEWrapper._signature_preamble
        second_part = MIMEApplication(_data=str(signature),
                                      _subtype=CustomMIMEWrapper._signature_subtype,
                                      _encoder=encode_7or8bit,
                                      name='signature.asc')
        second_part.add_header('Content-Description', 'OpenPGP digital signature')
        second_part.add_header('Content-Disposition', 'attachment', filename='signature.asc')
        self.msg.attach(second_part)
        return second_part

    def get_encrypted(self):
        try:
            msg = PGPMessage.from_blob(self.msg.get_payload(1).get_payload())
        except:
            return
        yield msg

    @staticmethod
    def copy_headers(from_msg, to_msg, overwrite=False):
        for key, value in from_msg.items():
            if overwrite:
                del to_msg[key]
            if key not in to_msg:
                to_msg[key] = value
        if to_msg.get_unixfrom() is None or overwrite:
            to_msg.set_unixfrom(from_msg.get_unixfrom())
        if (hasattr(from_msg, 'original_size')
                and (getattr(to_msg, 'original_size', None) is None
                     or overwrite)):
            to_msg.original_size = from_msg.original_size

    def decrypt(self, key):
        pmsg = next(iter(self.get_encrypted()))
        decrypted = key.decrypt(pmsg)

        dmsg = decrypted.message
        if isinstance(dmsg, bytearray):
            dmsg = dmsg.decode(decrypted.charset or 'utf-8')

        out = message_from_string(dmsg)
        if decrypted.is_signed:
            signature = next(iter(decrypted.signatures))
            self._wrap_signed(out, signature)
        else:
            self.msg.set_payload(out.get_payload())
            self.copy_headers(out, self.msg, True)
        return self

    def sign(self, key, **kwargs):
        payload = self.openpgp_mangle_for_signature(self.msg)

        signature = key.sign(payload, **kwargs)
        original_msg = copy.deepcopy(self.msg)
        return payload, self._wrap_signed(original_msg, signature)

    def _wrap_encrypted(self, payload):
        self.msg.set_payload([])
        self.msg.set_type(CustomMIMEWrapper._encrypted_multipart)
        self.msg.set_param('protocol', CustomMIMEWrapper._encrypted_type)
        self.msg.preamble = CustomMIMEWrapper._encryption_preamble
        first_part = MIMEApplication(_data='Version: 1',
                                     _subtype=CustomMIMEWrapper._encryption_subtype,
                                     _encoder=encode_7or8bit)
        first_part.add_header('Content-Description',
                              'PGP/MIME version identification')
        self.msg.attach(first_part)
        second_part = MIMEApplication(_data=str(payload),
                                      _subtype='octet-stream',
                                      _encoder=encode_7or8bit,
                                      name='encrypted.asc')
        second_part.add_header('Content-Description',
                               'OpenPGP encrypted message')
        second_part.add_header('Content-Disposition', 'inline',
                               filename='encrypted.asc')
        self.msg.attach(second_part)

    def _encrypt(self, pmsg, keys, cipher, **kwargs):
        emsg = copy.copy(pmsg)
        if len(keys) == 1:
            emsg = keys[0].encrypt(emsg, cipher=cipher, **kwargs)
        else:
            session_key = cipher.gen_key()
            for key in keys:
                emsg = key.encrypt(emsg, cipher=cipher,
                                   sessionkey=session_key,
                                   **kwargs)
            del session_key
        return emsg

    def encrypt(self, key, keys, **kwargs):
        hash = kwargs.get("hash", None)
        if not kwargs.get("cipher"):
            kwargs["cipher"] = SymmetricKeyAlgorithm.AES256

        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        payload = next(iter(self.get_payload()))
        pmsg = PGPMessage.new(payload)
        pmsg = self._encrypt(pmsg, keys, **kwargs)
        self._wrap_encrypted(pmsg)
        return self

    def sign_encrypt(self, key, keys, **kwargs):
        hash = kwargs.get("hash", None)

        if not kwargs.get("cipher"):
            kwargs["cipher"] = SymmetricKeyAlgorithm.AES256

        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        payload = next(iter(self.get_payload()))
        pmsg = PGPMessage.new(payload)
        pmsg |= key.sign(pmsg, hash=hash)
        pmsg = self._encrypt(pmsg, keys, **kwargs)
        self._wrap_encrypted(pmsg)
        return self

    def is_signed(self):
        if not self._is_mime():
            return False
        content_type = collapse_rfc2231_value(self.msg.get_payload(1).get_content_type())
        signed = content_type == CustomMIMEWrapper._signed_type or content_type == CustomMIMEWrapper._signed_multipart
        signed = signed and self.msg.get_content_subtype() == "signed"
        return signed

    def is_encrypted(self):
        if not self._is_mime():
            return False
        first_part = self.msg.get_payload(0).as_string()
        first_type = self.msg.get_payload(0).get_content_type()
        second_type = self.msg.get_payload(1).get_content_type()
        content_subtype = self.msg.get_content_subtype()
        protocol_param = collapse_rfc2231_value(self.msg.get_param('protocol', ''))
        return ('Version: 1' in first_part and
                first_type == CustomMIMEWrapper._encrypted_type and
                second_type == 'application/octet-stream' and
                content_subtype == 'encrypted' and
                protocol_param == CustomMIMEWrapper._encrypted_type)



from email_master.compat import base64_decode_to_bytes, base64_encode_bytes
from pgpy.constants import HashAlgorithm
from pgpy.errors import PGPError
from email_master.parser import EmailMasterParser
from email_master.pgp.wrapper import CustomMIMEWrapper

from pgpy import PGPKey, PGPSignature
from pgpy.types import SignatureVerification
import email
import tempfile

import six
import random
from itertools import combinations


class _PGPType(object):
    CLEARTEXT = "Cleartext"

    SIGNED = "Signed"
    ENCRYPTED = "Encrypted"

    SIGNED_AND_ENCRYPTED = "Signed & Encrypted"

    SIGNED_AND_VERIFIED = "Signed & Verified"
    SIGNED_AND_ENCRYPTED_AND_VERIFIED = "Signed & Encrypted & Verified"

    SIGNED_AND_NOT_VERIFIED = "Signed & Not Verified"
    SIGNED_AND_ENCRYPTED_NOT_VERIFIED = "Signed & Encrypted & Not Verified"


class PGPIngestType(object):
    def __init__(self, signed=False, encrypted=False):
        self.signed = signed
        self.encrypted = encrypted

    def __str__(self):
        s = _PGPType.CLEARTEXT  # 00

        if self.signed and self.encrypted:  # 11
            s = _PGPType.SIGNED_AND_ENCRYPTED
        elif self.signed:  # 10
            s = _PGPType.SIGNED
        elif self.encrypted: # 01
            s = _PGPType.ENCRYPTED
        return s

    def to_ingest_type(self):
        return self

    def to_output_type(self, verify=False):
        """
        Convert PGPType from Ingest type (which doesn't care about verification) to PGPOutputType (which does)
        """
        return PGPOutputType(self.signed, self.encrypted, verify)

    def needs_pkeys(self):
        """
        Returns whether using this configtype requires public/private keys to continue execution
        """
        if self.signed or self.encrypted:
            return True
        else:
            return False


class PGPOutputType(object):
    def __init__(self, signed=False, encrypted=False, verified=False):
        self.signed = signed
        self.encrypted = encrypted
        self.verified = verified
        str(self)  # Does validity check

    def __str__(self):
        if self.signed and self.encrypted and self.verified:  # 111
            s = _PGPType.SIGNED_AND_ENCRYPTED_AND_VERIFIED
        elif self.signed and self.encrypted and not self.verified:  # 110
            s = _PGPType.SIGNED_AND_ENCRYPTED_NOT_VERIFIED
        elif self.signed and not self.encrypted and self.verified:  # 101
            s = _PGPType.SIGNED_AND_VERIFIED
        elif self.signed and not self.encrypted and not self.verified:  # 100
            s = _PGPType.SIGNED
        elif not self.signed and self.encrypted and self.verified:  # 011
            raise ValueError("Verify cannot be True if the email is not signed!")
        elif not self.signed and self.encrypted and not self.verified:  # 010
            s = _PGPType.ENCRYPTED
        elif not self.signed and not self.encrypted and self.verified:  # 001
            raise ValueError("Verify cannot be True if the email is not signed!")
        elif not self.signed and not self.encrypted and not self.verified:  # 000
            s = _PGPType.CLEARTEXT
        else:
            raise ValueError("Invalid config!")  # Shouldn't be here ever

        return s

    def to_output_type(self, verify=False):
        return self

    def to_ingest_type(self):
        """
        Convert PGP Type from Output type (which cares about verification) to IngestType (which doesn't)
        """
        return PGPIngestType(self.signed, self.encrypted)

    def needs_pkeys(self):
        """
        Returns whether using this configtype requires public/private keys to continue execution
        """
        if self.signed or self.encrypted or self.verified:
            return True
        else:
            return False


class PGPConfig(object):
    def __undefined_attr(self, attr_name):
        def attr(self):
            raise ValueError("Property '{}' not defined and cannot be accessed".format(attr_name))

        return property(attr)

    def __init__(self, pgp_action=None, pgp_private_b64="", pgp_public_b64="", pgp_password=""):
        self.pgp_action = pgp_action
        self.key_pw = pgp_password

        ingest = isinstance(pgp_action, PGPIngestType)
        output = isinstance(pgp_action, PGPOutputType)
        provided_pkeys = not bool(pgp_private_b64 == "" and pgp_public_b64 == "")

        if not ingest and not output:
            raise ValueError("Invalid pgp_action, must be of type PGPIngestType or PGPOutputType!")

        if pgp_action.needs_pkeys() and not provided_pkeys:
            if ingest:
                raise ValueError("Must provide PGP private and public keys to sign/encrypt!")
            elif output:
                raise ValueError("Must provide PGP private and public keys to decrypt/verify signatures!")

        self.priv_key, self.priv_keystore = self._key_from_b64(pgp_private_b64)
        self.pub_key, self.pub_keystore = self._key_from_b64(pgp_public_b64)

        if self.priv_key:
            try:
                with self.priv_key.unlock(self.key_pw):
                    pass
            except Exception as e:
                raise ValueError("Invalid private key or password, key unlock failed! Error: '{}'".format(str(e)))
        else:
            setattr(PGPConfig, "priv_key", self.__undefined_attr("Private Key"))

        if not self.pub_key:
            setattr(PGPConfig, "pub_key", self.__undefined_attr("Public Key"))

    def _generate_bound(self):
        valid_chars = []
        valid_chars.extend(range(ord("A"), ord("Z")))  # A-Z
        valid_chars.extend(range(ord("a"), ord("z")))  # a-z
        valid_chars.extend(range(ord("0"), ord("9")))  # 0-9
        valid_chars = [chr(rr) for rr in valid_chars]  # Convert to chars
        if six.PY3:
            return "".join(random.choices(valid_chars, k=32))  # Sample 32 at random (py3 only)
        else:
            return "".join([random.choice(valid_chars) for _ in range(32)])  # py2 32 samples

    def _key_from_b64(self, data):
        """
        Return a PGPKey from base64 data, if no data return None
        """
        if not data:
            return None, None

        f = tempfile.NamedTemporaryFile(suffix=b'')
        dd = base64_decode_to_bytes(data)
        f.write(dd)
        f.seek(0)
        key, keystore = PGPKey.from_file(f.name)
        f.close()
        return key, keystore

    def _find_duplicated_newlines(self, msg_str, find_char='\r'):
        """
        Find the indexes of duplicate carriage returns, which mess up pgp verification
        Return a list of message strings with modified bodies

        Example: (newlines have been removed
            ...
            Content-Disposition: attachment; filename="Test.eml"\r
            MIME-Version: 1.0\r
            \r
            \r
            <base64>

        The double '\r's (technically triple) was added by python's email parsing package, since it ingested the attachment as a message obj,
        instead of an attachment object, and automatically adds a newline before message objects after using .to_string()
        so it must be removed.

        But we can't be sure that the double \r wasn't just an artifact of the email (since we're doing a global str search)
        so we need to grab all iterations of the double newline.
        So for example say we process an email and find the following indexes have double \r's
        [(31, 32), (45, 46), (78,79)] -> lets call these [A, B, C]
        So now we want all bodies with any combination of these replacements (as any replacement could be real, or an error)
        Essentially -> [[()], [('A',), ('B',), ('C',)], [('A', 'B'), ('A', 'C'), ('B', 'C')], [('A', 'B', 'C')]]
        and then turn those replacements into different bodies.
        Since this is an exponentially increasing function, we don't want to pre-calculate every combination, since
        we break on the first combination that works, so we use generators
        """

        msg_list = msg_str.split("\n")
        idxs = []
        for i in range(len(msg_list)):
            if i + 1 >= len(msg_list):
                continue
            if msg_list[i] == find_char and msg_list[i] == msg_list[i + 1]:
                idxs.append(i)  # Add index that needs to be removed

        num_idxs = len(idxs)
        all_combins = [list(combinations(idxs, size)) for size in range(num_idxs + 1)]  # list of all combinations
        combins = []
        [combins.extend(c) for c in all_combins]  # combine all combinations

        # Unique value to substitute in the list to perform multi replacements without damaging the list due to shifting
        remove_obj = object()

        for comb in combins:
            new_msg_str = str(msg_str)
            new_msg_list = new_msg_str.split("\n")
            for c in comb:
                new_msg_list[c] = remove_obj

            new_msg_list = list(filter(lambda x: x != remove_obj, new_msg_list))  # Filter out the removed objs
            g = "\n".join(new_msg_list)  # rejoin the list
            yield g

    def _find_strip_newlines(self, msg_str):
        """
        Function to find when newlines with whitespace are being duplicated after send/ingest, seems to happen with different mail
        clients. Returns a generator for possible fixes to verify PGP sig
        """
        func_list = [  # List of functions to apply to each line of the msg
            lambda x: x.rstrip(),
            lambda x: x.lstrip(),
            lambda x: x.strip()
        ]

        for stripped_msg in self._find_duplicated_newlines(msg_str, find_char=""):
            for func in func_list:
                yield "\n".join(list(map(func, stripped_msg.split("\n"))))

        # self.pub_key.verify("\n".join(list(map(lambda x: x.rstrip(), aa.split("\n")))), pgp_blob)

    def _bodies_to_test(self, msg_str):
        """
        Create a list of message strings that have been modified slightly (due to poor python parsing) to
        recreate the original email bytes exactly as sent, to verify the PGP signature
        """
        # Bodies is a list of lists, where the inner lists can be generators
        bodies = [
            [
                msg_str,  # Start with unmodified body
                msg_str.replace("\n", "\r\n")  # Need to replace newlines with \r\n "sometimes"
            ],
            self._find_duplicated_newlines(msg_str.replace("\n", "\r\n")),  # Generator for replacements for newline fix
            self._find_duplicated_newlines(msg_str, find_char=''),
            self._find_strip_newlines(msg_str),  # Generator for extra whitespace fix
            self._find_duplicated_newlines(msg_str.replace("\n", "\r\n"), find_char=''),
            self._find_duplicated_newlines(msg_str)  # Generator for newline fix
        ]
        for body_list in bodies:
            for body in body_list:
                yield body

    def _find_pgp_sig(self, msg_obj, enc_body):
        attachments = len(msg_obj.get_payload())
        for idx in range(attachments):
            try:
                pgp_blob = msg_obj.get_payload(idx).get_payload()
                pgp_blob = PGPSignature.from_blob(pgp_blob)

                msg_str = msg_obj.get_payload(0).as_string()  # Hard coded 0 for base payload to verify against
                bodies = self._bodies_to_test(msg_str)
                is_pgp_verified = False  # Did we find a body that verified?
                for body in bodies:
                    is_pgp_verified = self.pub_key.verify(body, pgp_blob)
                    if bool(is_pgp_verified):  # This body worked
                        break
                if is_pgp_verified and isinstance(is_pgp_verified, SignatureVerification):
                    return list(is_pgp_verified.good_signatures)
                else:
                    return []
            except PGPError as e:
                continue  # Invalid signature
            except TypeError as e:
                continue  # Invalid body type
            except ValueError as e:
                continue  # Invalid body type (for encrypted message)
        return []  # Couldn't find an attachment with a valid signature

    def _check_pgp_verification(self, mime_msg, message_type, fingerprints, enc_body=None):
        """
        Check the verification status of a signature of a message, returning the message type
        and fingerprints.

        Function is needed because messages can be encrypted before or after the signature is created, leading to the
        need to check it twice (call this func twice)
        """
        sigs = self._find_pgp_sig(mime_msg.msg, enc_body)
        is_verified = len(sigs) > 0  # In case we're already verified

        message_type.verified = message_type.verified or is_verified
        fingerprints.extend([str(sig.by.fingerprint) for sig in sigs])

        return message_type, fingerprints

    def unlock(self, msg_as_string, parser_kwargs):
        # TODO Put check if we should even be able to unlock, raise exception upon method call instead of
        # attribute access. Also would be able to not fail upon being unable to unlock an email

        mime_msg = CustomMIMEWrapper(msg_as_string)
        message_type = PGPOutputType()
        fingerprints = []
        warnings = "None"

        if EmailMasterParser(base64_encode_bytes(str(mime_msg.msg).encode()), **parser_kwargs).parse()["subject"].startswith("Test 1d"):
            tw = 2

        message_type.encrypted = mime_msg.is_encrypted()
        message_type.signed = mime_msg.is_signed()  # Check if the message is signed before decrypting

        # Check pgp verification before decrypting
        message_type, fingerprints = self._check_pgp_verification(mime_msg, message_type, fingerprints)
        encrypted_body = None

        if message_type.encrypted:
            if len(mime_msg.msg.get_payload()) == 2:  # One part pgp header, one part pgp encrypted body
                # TODO Multiple encrypted bodies?
                encrypted_body = mime_msg.msg.get_payload(1).as_string()

            with self.priv_key.unlock(self.key_pw) as unlocked_key:
                try:
                    decrypted = mime_msg.decrypt(unlocked_key)
                    if not decrypted:
                        raise ValueError("No decrypted message returned from mime_msg.decrypt!")
                    mime_msg = decrypted
                except PGPError as e:
                    # This means the message cannot be decrypted with this signature
                    warnings = "Attempted decrypt failed, '{}'".format(str(e))

        # Check if the message is signed after decrypting
        message_type.signed = mime_msg.is_signed() or message_type.signed

        # Check pgp verification after decrypting
        message_type, fingerprints = self._check_pgp_verification(mime_msg, message_type, fingerprints, encrypted_body)

        if len(mime_msg.msg._payload) == 2 and message_type.signed:
            mime_msg.msg._payload[0]._headers.extend(
                mime_msg.msg._headers)  # Copy outer headers into the inner message object
            mime_msg.msg._payload[0].preamble = mime_msg.msg.preamble  # Copy the preamble

            orig_msg = EmailMasterParser(base64_encode_bytes(str(mime_msg.msg.get_payload(0)).encode()),
                                         **parser_kwargs).parse()  # Orig message
            sig_msg = EmailMasterParser(base64_encode_bytes(str(mime_msg.msg.get_payload(1)).encode()),
                                        **parser_kwargs).parse()  # Signature
            merge_keys = ["attachments_sha1", "attachments_md5", "attachments_sha256", "attach_info", "attachments"]
            for k in merge_keys:
                if isinstance(orig_msg[k], list):
                    orig_msg[k].extend(sig_msg[k])
                else:
                    ks = orig_msg[k].split(",")
                    if ks == ['']:
                        ks = []
                    ks.append(sig_msg[k])
                    orig_msg[k] = ",".join(ks)

            email_data = orig_msg
        else:
            email_data = EmailMasterParser(base64_encode_bytes(str(mime_msg.msg).encode()), **parser_kwargs).parse()

        email_data.update({
            "message_type": str(message_type),
            "is_verified": message_type.verified,
            "is_signed": message_type.signed,
            "is_encrypted": message_type.encrypted,
            "fingerprints": fingerprints,
            "warnings": warnings
        })
        return email_data

    def _format_signed_payload(self, mime_message, payload, signed_payload):
        bound = self._generate_bound()
        new_headers = list(filter(lambda x: x[0] != "Content-Type", email.message_from_string(payload)._headers))

        signed_payload = signed_payload.as_string().replace("\n", "\r\n")

        header_data = "\r\n".join(["{}: {}".format(b[0], b[1]) for b in new_headers])
        newmsg = header_data + \
                 "\r\nContent-Type: multipart/signed; micalg=pgp-sha256; " + \
                 "protocol=\"application/pgp-signature\";" + \
                 "boundary=\"{bound}\"\r\n--{bound}\r\n".format(bound=bound) + \
                 payload + \
                 "\r\n--{bound}\r\n".format(bound=bound) + \
                 signed_payload + \
                 "--{}--".format(bound)
        return newmsg

    def lock(self, mime_message):
        if self.pgp_action.signed and self.pgp_action.encrypted:
            with self.priv_key.unlock(self.key_pw) as k:
                mime = mime_message.sign_encrypt(k, list(self.pub_keystore.values()))
                return str(mime.msg)
        elif self.pgp_action.encrypted:
            with self.priv_key.unlock(self.key_pw) as k:
                mime = mime_message.encrypt(k, list(self.pub_keystore.values()))
                return str(mime.msg)
        elif self.pgp_action.signed:
            with self.priv_key.unlock(self.key_pw) as k:
                payload, signed_payload = mime_message.sign(k, hash=HashAlgorithm.SHA256)
                return self._format_signed_payload(mime_message, payload, signed_payload)
        else:
            return mime_message.msg.as_string()



import hashlib
import glob
from email_master.compat import base64_encode, to_unicode

import six
if six.PY3:
    unicode = to_unicode


class EmailAttachmentList(object):
    """Container class for attachments, to standardize the output"""

    def __init__(self):
        self.attachments = []

    def filter_by_filename(self, exclude=None, include=None):
        """
        Filenames should be a tuple of endings like (".png", ".gif", ..) to remove from the attachment list
        Returns a new EmailAttachmentList instance
        """

        def do_filter(attch):
            # Only return included if provided. Else do excluded if provided.
            if include:  # Filter out attachments that DO NOT end with any in the list
                return bool(attch.filename.lower().endswith(include))
            elif exclude:
                # Filter out attachments that end with any in the list
                return not bool(attch.filename.lower().endswith(exclude))

        eal = EmailAttachmentList()
        eal.attachments = list(filter(do_filter, self.attachments))
        return eal

    def add_attachment(self, email_attachment):
        if isinstance(email_attachment, EmailAttachment):
            self.attachments.append(email_attachment)
        else:
            raise Exception("Attempted to add attachment to EmailAttachments, invalid type detected")

    def to_swimlane_output(self):
        result = {
            "attach_info": [],
            "attachments": [],
            "attachments_md5": [],
            "attachments_sha1": [],
            "attachments_sha256": []
        }

        # Add all attachments into a list
        for attachment in self.attachments:
            result["attachments"].append(attachment.attachment_data)
            result["attach_info"].append(attachment.header_info)
            result["attachments_sha1"].append(attachment.hash_sha1)
            result["attachments_md5"].append(attachment.hash_md5)
            result["attachments_sha256"].append(attachment.hash_sha256)

        # Flatten results
        for k, v in result.items():
            if k != "attachments":  # Don't flatten the attachments
                result[k] = ",".join(result[k])

        return result


class EmailAttachment(object):
    """Singular Attachment"""

    def __init__(self, header_info, filename, raw_data):
        """
        Create a singular email attachment object
        :param header_info: String of header information about the email
        :param filename: Filename of the attachment
        :param raw_data: Byte-like data, will be base64encoded
        """
        self.header_info = header_info or u""
        self.raw_data = raw_data or u""
        self.filename = filename or u""

        if six.PY3 and not isinstance(self.raw_data, bytes):
            self.raw_data = self.raw_data.encode()

        self.hash_md5 = hashlib.md5(self.raw_data).hexdigest()
        self.hash_sha1 = hashlib.sha1(self.raw_data).hexdigest()
        self.hash_sha256 = hashlib.sha256(self.raw_data).hexdigest()
        self.attachment_data = {
            "filename": self.filename,
            "base64": base64_encode(self.raw_data)
        }


import six
import base64

EmailMessage = None
MIMEText = None
MIMEMultipart = None

if six.PY3:
    from email.mime.text import MIMEText
    from email.message import EmailMessage
    from email.mime.multipart import MIMEMultipart
else:
    from email.MIMEText import MIMEText
    from email.MIMEMultipart import MIMEMultipart
    from email.message import Message as EmailMessage


def base64_encode_bytes(data, *args, **kwargs):
    return base64.b64encode(data, *args, **kwargs).decode("ascii")


def base64_encode(data, *args, **kwargs):
    if isinstance(data, bytes):
        return base64_encode_bytes(data, *args, **kwargs)

    if six.PY3:
        input_data = data.encode("utf-8")
    else:
        input_data = data

    output_data = base64.b64encode(input_data, *args, **kwargs)
    if six.PY3:
        return output_data.decode("ascii")
    else:
        return output_data


def base64_decode_to_bytes(data, *args, **kwargs):
    return base64.b64decode(data, *args, **kwargs)


def base64_decode(data, *args, **kwargs):
    output_data = base64_decode_to_bytes(data, *args, **kwargs)
    if six.PY3:
        return output_data.decode()
    else:
        return output_data


def to_unicode(ss, enc=None, errors=None):
    # Strings are already unicode inheritly in python3
    return ss
